# Pokročilý Python

## Iterace

- `reversed`
    - totéž lze jen pomocí `range`, ale je to nepraktické
- `enumerate`
    - motivace: `for x in iterable` je super, ale někdy jsou potřeba indexy
        - pořád chceme iterovat přes prvky, jenom přidat indexy navíc
        - v Pythonu je zvykem to dělat takto, preferovanější než `for i in range(len(iterable))`
    - využití: přičtení seznamu k druhému, vypisování prvočísel u Eratosthenova síta
- `zip`
    - podobná motivace, nechceme indexy
    - využití: CSV, dvě sousední hodnoty
- list comprehensions
    - využití: transformace seznamu, filtrování, rozbalení seznamu seznamů
        - obecně oneliners, dobře se čtou -- říkám, co chci, ne jak toho docílit
    - pomocí `map` a `filter` jde často udělat totéž
    - hodí se pro inicializaci seznamů, protože `*` jenom zkopíruje referenci
- obdobně dict comprehensions
    - často ve spojení s `enumerate` nebo `zip`

## Dekorátory

- `@staticmethod`
    - vypne automatické nastavování prvního argumentu
- `@classmethod`
    - více "konstruktorů"
- `@dataclass`
    - automatický init, vypisování, rovnost, hash
- `@lru_cache`
    - automatické cachování výsledků funkce
- je možné zadefinovat vlastní, některé knihovny to dělají

## Typing

- první motivace: napovídání v IDE
- typy některých proměnných lze odvodit automaticky
    - někdy to nejde: argumenty a návratové hodnoty funkcí, seznamy
- type annotations jako řešení
    - při vykonávání kódu nemají význam
- speciální typy: `List`, `Tuple`, `Optional`, `Callable`
- položka stejného typu jako vnější třída: do uvozovek nebo `from __future__ import annotations`
- kromě napovídání také typechecking: `mypy`, `mypy --strict`
- ukázka: klient pro strategickou hru ze sousu

## Moduly

- mnohé dostupné jako součást standardní knihovny
- nespočet dalších externě
- ukázka: `requests`
    - `pip3 install requests`
- virtual environment
    - `python3 -m venv venv`
    - `. venv/bin/activate`
    - proč? může vyjít nová verze balíčku, která můj kód rozbije, proto chci oddělené prostředí
    - také se může stát, že některé balíčky budou mít mezi sebou konflikt
- modelová situace: najdu na Githubu cool projekt, stáhnu si jej, chci jej spustit, potřebuje balíčky
    - jak zjistím, které balíčky potřebuju? => `requirements.txt`
        - `pip3 install -r requirements.txt`
    - před instalací samozřejmě `venv`
